"""Copyright (C) 2024 TheOnlyWayUp

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/.

---

Entrypoint."""

from wattpad.wattpad import User, Story, List



"""Copyright (C) 2024 TheOnlyWayUp

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/.

---

Pydantic Models representing Wattpad API Responses. Thanks https://jsontopydantic.com."""

from __future__ import annotations


from typing import List, Optional
from pydantic import BaseModel, Field


class InboxModel(BaseModel):
    """Represents a User's Inbox."""

    unread: int
    total: int


class NotificationsModel(BaseModel):
    """Represents a User's Notifications."""

    unread: int


class ConnectedServicesModel(BaseModel):
    """Represents a User's Connected Services."""

    facebook: bool
    twitter: bool


class UserModel(BaseModel):
    """Represents a User."""

    username: str

    avatar: Optional[str] = None
    is_private: Optional[bool] = Field(None, alias="isPrivate")
    background_url: Optional[str] = Field(None, alias="backgroundUrl")
    follower: Optional[bool] = None
    following: Optional[bool] = None
    name: Optional[str] = None
    description: Optional[str] = None
    status: Optional[str] = None
    gender: Optional[str] = None
    gender_code: Optional[str] = Field(None, alias="genderCode")
    language: Optional[int] = None
    locale: Optional[str] = None
    create_date: Optional[str] = Field(None, alias="createDate")
    modify_date: Optional[str] = Field(None, alias="modifyDate")
    location: Optional[str] = None
    verified: Optional[bool] = None
    ambassador: Optional[bool] = None
    facebook: Optional[str] = None
    twitter: Optional[str] = None
    website: Optional[str] = None
    lulu: Optional[str] = None
    smashwords: Optional[str] = None
    bubok: Optional[str] = None
    votes_received: Optional[int] = Field(None, alias="votesReceived")
    num_stories_published: Optional[int] = Field(None, alias="numStoriesPublished")
    num_following: Optional[int] = Field(None, alias="numFollowing")
    num_followers: Optional[int] = Field(None, alias="numFollowers")
    num_messages: Optional[int] = Field(None, alias="numMessages")
    num_lists: Optional[int] = Field(None, alias="numLists")
    verified_email: Optional[bool] = None
    preferred_categories: Optional[List] = None
    allow_crawler: Optional[bool] = Field(None, alias="allowCrawler")
    deeplink: Optional[str] = None
    is_muted: Optional[bool] = Field(None, alias="isMuted")
    birthdate: Optional[str] = None
    inbox: Optional[InboxModel] = None
    notifications: Optional[NotificationsModel] = None
    connected_services: Optional[ConnectedServicesModel] = Field(
        None, alias="connectedServices"
    )
    age: Optional[int] = None
    email: Optional[str] = None
    has_password: Optional[bool] = None


class LanguageModel(BaseModel):
    """Represents a Language."""

    id: int
    name: str


class FirstPublishedPartModel(BaseModel):
    """Represents the first published part of a Story."""

    id: int
    create_date: str = Field(..., alias="createDate")


class LastPublishedPartModel(BaseModel):
    """Represents the last (most recent) published part of a Story."""

    id: int
    create_date: str = Field(..., alias="createDate")


class PartModel(BaseModel):
    """Represents a Part of a Story."""

    id: int
    title: str
    url: str
    modify_date: str = Field(..., alias="modifyDate")
    create_date: str = Field(..., alias="createDate")
    comment_count: int = Field(..., alias="commentCount")
    vote_count: int = Field(..., alias="voteCount")
    read_count: int = Field(..., alias="readCount")


class TagRankingModel(BaseModel):
    """Represents a Story's Tag Rankings."""

    name: str
    rank: int
    total: int


class StoryModel(BaseModel):
    """Represents a Story."""

    id: str

    title: Optional[str] = None
    create_date: Optional[str] = Field(None, alias="createDate")
    modify_date: Optional[str] = Field(None, alias="modifyDate")
    vote_count: Optional[int] = Field(None, alias="voteCount")
    read_count: Optional[int] = Field(None, alias="readCount")
    comment_count: Optional[int] = Field(None, alias="commentCount")
    description: Optional[str] = None
    completed: Optional[bool] = None
    tags: Optional[List[str]] = None
    rating: Optional[int] = None
    mature: Optional[bool] = None
    url: Optional[str] = None
    is_paywalled: Optional[bool] = Field(None, alias="isPaywalled")
    cover: Optional[str] = None
    cover_timestamp: Optional[str] = None
    categories: Optional[List[int]] = None
    copyright: Optional[int] = None
    first_part_id: Optional[int] = Field(None, alias="firstPartId")
    num_parts: Optional[int] = Field(None, alias="numParts")
    deleted: Optional[bool] = None

    first_published_part: Optional[FirstPublishedPartModel] = Field(
        None, alias="firstPublishedPart"
    )
    last_published_part: Optional[LastPublishedPartModel] = Field(
        None, alias="lastPublishedPart"
    )
    language: Optional[LanguageModel] = None
    user: Optional[UserModel] = None
    parts: List[PartModel] = []
    tag_rankings: list[TagRankingModel] = Field([], alias="tagRankings")


class ListModel(BaseModel):
    """Represents a User's List."""

    id: int
    name: str
    stories: List[StoryModel]



"""Copyright (C) 2024 TheOnlyWayUp

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/.

---

Typehints for the models in `./models.py`. These types are used to specify what fields to request in API-Interfacing methods."""

from __future__ import annotations

from typing import TypedDict
from typing_extensions import NotRequired


class InboxModelFieldsType(TypedDict):
    """Typehints for the Inbox model."""

    unread: NotRequired[bool]
    total: NotRequired[bool]


class NotificationsModelFieldsType(TypedDict):
    """Typehints for the Notifications model."""

    unread: NotRequired[bool]


class ConnectedServicesModelFieldsType(TypedDict):
    """Typehints for the Notifications model."""

    facebook: NotRequired[bool]
    twitter: NotRequired[bool]


class UserModelFieldsType(TypedDict):
    """Typehints for the User model."""

    username: NotRequired[bool]
    avatar: NotRequired[bool]
    isPrivate: NotRequired[bool]
    backgroundUrl: NotRequired[bool]
    follower: NotRequired[bool]
    following: NotRequired[bool]
    name: NotRequired[bool]
    description: NotRequired[bool]
    status: NotRequired[bool]
    gender: NotRequired[bool]
    genderCode: NotRequired[bool]
    language: NotRequired[bool]
    locale: NotRequired[bool]
    createDate: NotRequired[bool]
    modifyDate: NotRequired[bool]
    location: NotRequired[bool]
    verified: NotRequired[bool]
    ambassador: NotRequired[bool]
    facebook: NotRequired[bool]
    twitter: NotRequired[bool]
    website: NotRequired[bool]
    lulu: NotRequired[bool]
    smashwords: NotRequired[bool]
    bubok: NotRequired[bool]
    votesReceived: NotRequired[bool]
    numStoriesPublished: NotRequired[bool]
    numFollowing: NotRequired[bool]
    numFollowers: NotRequired[bool]
    numMessages: NotRequired[bool]
    numLists: NotRequired[bool]
    verified_email: NotRequired[bool]
    preferred_categories: NotRequired[bool]
    allowCrawler: NotRequired[bool]
    deeplink: NotRequired[bool]
    isMuted: NotRequired[bool]
    birthdate: NotRequired[bool]
    age: NotRequired[int]
    email: NotRequired[bool]
    has_password: NotRequired[bool]

    inbox: NotRequired[InboxModelFieldsType | bool]
    notifications: NotRequired[NotificationsModelFieldsType | bool]
    connectedServices: NotRequired[ConnectedServicesModelFieldsType | bool]


class LanguageModelFieldsType(TypedDict):
    """Typehints for the Language model."""

    id: NotRequired[bool]
    name: NotRequired[bool]


class PublishedModelFieldsType(TypedDict):
    """Typehints for the Published model (for the firstPublishedPart and lastPublishedPart fields)."""

    id: NotRequired[bool]
    createDate: NotRequired[bool]


class PartModelFieldsType(TypedDict):
    """Typehints for the Part model."""

    id: NotRequired[bool]
    title: NotRequired[bool]
    url: NotRequired[bool]
    modifyDate: NotRequired[bool]
    createDate: NotRequired[bool]
    commentCount: NotRequired[bool]
    voteCount: NotRequired[bool]
    readCount: NotRequired[bool]


class TagRankingModelFieldsType(TypedDict):
    """Typehints for the TagRankings model."""

    name: NotRequired[bool]
    rank: NotRequired[bool]
    total: NotRequired[bool]


class StoryModelFieldsType(TypedDict):
    """Typehints for the Story model."""

    id: NotRequired[bool]
    title: NotRequired[bool]
    createDate: NotRequired[bool]
    modifyDate: NotRequired[bool]
    voteCount: NotRequired[bool]
    readCount: NotRequired[bool]
    commentCount: NotRequired[bool]
    description: NotRequired[bool]
    completed: NotRequired[bool]
    tags: NotRequired[bool]
    rating: NotRequired[bool]
    mature: NotRequired[bool]
    url: NotRequired[bool]
    isPaywalled: NotRequired[bool]
    cover: NotRequired[bool]
    cover_timestamp: NotRequired[bool]
    categories: NotRequired[bool]
    copyright: NotRequired[bool]
    firstPartId: NotRequired[bool]
    numParts: NotRequired[bool]
    deleted: NotRequired[bool]

    parts: NotRequired[PartModelFieldsType | bool]
    tagRankings: NotRequired[TagRankingModelFieldsType | bool]
    firstPublishedPart: NotRequired[PublishedModelFieldsType | bool]
    lastPublishedPart: NotRequired[PublishedModelFieldsType | bool]
    language: NotRequired[LanguageModelFieldsType | bool]
    user: NotRequired[UserModelFieldsType | bool]


class ListModelFieldsType(TypedDict):
    """Typehints for the List model."""

    id: NotRequired[bool]
    name: NotRequired[bool]

    stories: NotRequired[StoryModelFieldsType | bool]



"""Copyright (C) 2024 TheOnlyWayUp

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/.

---

The main module for the wattpad package. This contains the User, Story, and List classes.

>>> u = User("<username>")
>>> await u.fetch()
>>> print(u.data)

Use `help(User)` for information. Refer to the documentation for more.
"""

from __future__ import annotations
from typing import Optional, cast
from .models import (
    ListModel,
    StoryModel,
    UserModel,
)
from .model_types import ListModelFieldsType, UserModelFieldsType, StoryModelFieldsType
from .utils import get_fields, build_url, fetch_url, construct_fields, singleton


@singleton
class User:
    """A representation of a User on Wattpad.
    **Note**: Users are singletons, unique as per their username. Two user classes with the same username are the _same_.

    Attributes:
        username (str): Lowercased username of this User.
        stories (list[Story]): Stories authored by this User.
        followers (list[User]): Users that follow this User.
        following (list[User]): Users this User follows.
        lists (list[List]): Lists created by this User.
        data (UserModel): User Data from the Wattpad API.
    """

    def __init__(self, username: str, **kwargs):
        """Create a User object.

        Args:
            username (str): The username of this User.
            **kwargs (any): Arguments to pass directly to the underlying `UserModel`. These are ignored if the User has been instantiated earlier in the runtime.
        """
        self.username = username.lower()
        self.stories: list[Story] = []
        self.followers: list[User] = []
        self.following: list[User] = []
        self.lists: list[List] = []

        self.data = UserModel(username=self.username, **kwargs)

    def __repr__(self) -> str:
        return f"<User username={self.username}>"

    async def fetch(self, include: bool | UserModelFieldsType = False) -> dict:
        """Populates a User's data. Call this method after instantiation.

        Args:
            include (bool | UserModelFieldsType, optional): Fields to fetch. True fetches all fields. Defaults to False.

        Returns:
            dict: The raw API Response.
        """
        if include is False:
            include_fields: UserModelFieldsType = {}
        elif include is True:
            include_fields: UserModelFieldsType = {
                key: True for key in get_fields(UserModel)  # type: ignore
            }
        else:
            include_fields: UserModelFieldsType = include

        data = cast(
            dict,
            await fetch_url(
                build_url(f"users/{self.data.username}", fields=dict(include_fields))
            ),
        )
        if "username" in data:
            data.pop("username")

        self.data = UserModel(username=self.username, **data)

        return data

    async def fetch_stories(self, include: bool | StoryModelFieldsType = False) -> dict:
        """Fetch a User's authored stories.

        Args:
            include (bool | StoryModelFieldsType, optional): Fields of authored stories to fetch. True fetches all fields. Defaults to False.

        Returns:
            dict: The raw API Response.
        """
        if include is False:
            include_fields: StoryModelFieldsType = {}
        elif include is True:
            include_fields: StoryModelFieldsType = {
                key: True for key in get_fields(StoryModel)  # type: ignore
            }
        else:
            include_fields: StoryModelFieldsType = include

        include_fields["id"] = True

        url = (
            build_url(f"users/{self.data.username}/stories", fields=None)
            + f"&fields=stories({construct_fields(dict(include_fields))})"  # ! The field format for story retrieval differs here. It's /stories?fields=stories(<fields>). Compared to the usual /path?fields=<fields>. For this reason, we need to manually edit the fields in.
        )
        data = cast(dict, await fetch_url(url))

        stories: list[Story] = []
        for story in data["stories"]:
            if "user" in story:
                story.pop("user")
            stories.append(Story(user=self, **story))

        self.stories = stories
        self.data.num_stories_published = len(
            self.stories
        )  # ! The data['total'] can also be used, but it isn't always present. (Based on included_fields.)

        return data

    async def fetch_followers(
        self,
        include: bool | UserModelFieldsType = False,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> dict:
        """Fetches the User's followers.

        Args:
            include (bool | UserModelFieldsType, optional): Fields of the following users' to fetch. True fetches all fields. Defaults to False.
            limit (Optional[int], optional): Maximum number of users to return at once. Use this alongside `offset` for better performance. Defaults to None.
            offset (Optional[int], optional): Number of users to skip before returning followers. Use this alongside `limit` for better performance. Defaults to None.

        Returns:
            dict: The raw API Response.
        """
        if include is False:
            include_fields: UserModelFieldsType = {}
        elif include is True:
            include_fields: UserModelFieldsType = {
                key: True for key in get_fields(UserModel)  # type: ignore
            }
        else:
            include_fields: UserModelFieldsType = include

        include_fields["username"] = True

        url = (
            build_url(
                f"users/{self.data.username}/followers",
                fields=None,
                limit=limit,
                offset=offset,
            )
            + f"&fields=users({construct_fields(dict(include_fields))})"  # ! Similar to story retrieval, requested fields need to be wrapped in `users(<fields>)`.
        )
        data = cast(dict, await fetch_url(url))

        self.followers = [User(**user) for user in data["users"]]
        self.data.num_followers = len(self.followers)

        return data

    async def fetch_following(
        self,
        include: bool | UserModelFieldsType = False,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> dict:
        """Fetch the users this User follows.

        Args:
            include (bool | UserModelFieldsType, optional): Fields of the followed users' to fetch. True fetches all fields. Defaults to False.
            limit (Optional[int], optional): Maximum number of users to return at once. Use this alongside `offset` for better performance. Defaults to None.
            offset (Optional[int], optional): Number of users to skip before returning followers. Use this alongside `limit` for better performance. Defaults to None.

        Returns:
            dict: The raw API Response.
        """
        if include is False:
            include_fields: UserModelFieldsType = {}
        elif include is True:
            include_fields: UserModelFieldsType = {
                key: True for key in get_fields(UserModel)  # type: ignore
            }
        else:
            include_fields: UserModelFieldsType = include

        include_fields["username"] = True

        url = (
            build_url(
                f"users/{self.data.username}/following",
                fields=None,
                limit=limit,
                offset=offset,
            )
            + f"&fields=users({construct_fields(dict(include_fields))})"  # ! Similar to story retrieval, requested fields need to be wrapped in `users(<fields>)`.
        )
        data = cast(dict, await fetch_url(url))

        self.following = [User(**user) for user in data["users"]]
        self.data.num_following = len(self.following)

        return data

    async def fetch_lists(
        self,
        include: bool | ListModelFieldsType = False,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> dict:
        """Fetch a User's lists.

        Args:
            include (bool | ListModelFieldsType, optional): Fields of the lists to fetch. True fetches all fields. Defaults to False.
            limit (Optional[int], optional): Maximum number of users to return at once. Use this alongside `offset` for better performance. Defaults to None.
            offset (Optional[int], optional): Number of users to skip before returning followers. Use this alongside `limit` for better performance. Defaults to None.

        Returns:
            dict: The raw API Response.
        """
        if include is False:
            include_fields: ListModelFieldsType = {}
        elif include is True:
            include_fields: ListModelFieldsType = {
                key: True for key in get_fields(ListModel)  # type: ignore
            }
        else:
            include_fields: ListModelFieldsType = include

        include_fields["id"] = True

        url = (
            build_url(
                f"users/{self.data.username}/lists",
                fields=None,
                limit=limit,
                offset=offset,
            )
            + f"&fields=lists({construct_fields(dict(include_fields))})"  # ! Similar to story retrieval, requested fields need to be wrapped in `lists(<fields>)`.
        )
        data = cast(dict, await fetch_url(url))

        self.lists = [List(user=self, **list_) for list_ in data["lists"]]
        self.data.num_lists = len(self.lists)

        return data


# --- #


@singleton
class Story:
    """A representation of a Story on Wattpad.
    **Note**: Stories are singletons, unique as per their ID. Two story classes with the same ID are the _same_.

    Attributes:
        id (str): Lowercased ID of this Story.
        user (User): The User who authored this Story.
        recommended (list[Story]): Stories recommended from this Story.
        data (StoryModel): Story Data from the Wattpad API.
    """

    def __init__(self, id: str, user: User, **kwargs):
        """Create a Story object.

        Args:
            id (str): The ID of the Story.
            user (User): The User who authored this Story.
        """
        self.id = id.lower()
        self.user: User = user
        self.recommended: list[Story] = []
        # self.parts: list[Part]  # ! NotImplemented. In the future, if Part text retrieval is a part of this library, that would warrant the creation of a seperate Part singleton. Right now, having self.parts would cause inconsistency with the rest of the library.
        self.data = StoryModel(id=self.id, **kwargs)

    def __repr__(self) -> str:
        return f"<Story id={self.id}>"

    async def fetch(self, include: bool | StoryModelFieldsType = False) -> dict:
        """Populates a Story's data. Call this method after instantiation.

        Args:
            include (bool | StoryModelFieldsType, optional): Fields to fetch. True fetches all fields. Defaults to False.

        Returns:
            dict: The raw API Response.
        """
        if include is False:
            include_fields: StoryModelFieldsType = {}
        elif include is True:
            include_fields: StoryModelFieldsType = {
                key: True for key in get_fields(StoryModel)  # type: ignore
            }
        else:
            include_fields: StoryModelFieldsType = include

        if "user" in include_fields:
            if include_fields["user"] is True:
                include_fields["user"] = cast(
                    UserModelFieldsType, {key: True for key in get_fields(UserModel)}  # type: ignore
                )
            elif include_fields["user"] is False:
                include_fields["user"] = {"username": True}
            else:
                include_fields["user"]["username"] = True
        else:
            include_fields["user"] = {"username": True}

        data = cast(
            dict,
            await fetch_url(
                build_url(f"stories/{self.data.id}", fields=dict(include_fields))
            ),
        )
        if "id" in data:
            data.pop("id")

        self.data = StoryModel(id=self.id, **data)

        return data

    async def fetch_recommended(
        self,
        include: bool | StoryModelFieldsType = False,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> list:
        """Fetch Stories recommended from this Story.

        Args:
            include (bool | StoryModelFieldsType, optional): Fields to fetch of the recommended stories. True fetches all fields. Defaults to False.
            limit (Optional[int], optional): Maximum number of users to return at once. Use this alongside `offset` for better performance. Defaults to None.
            offset (Optional[int], optional): Number of users to skip before returning followers. Use this alongside `limit` for better performance. Defaults to None.

        Returns:
            dict: The raw API Response.
        """
        if include is False:
            include_fields: StoryModelFieldsType = {}
        elif include is True:
            include_fields: StoryModelFieldsType = {
                key: True for key in get_fields(StoryModel)  # type: ignore
            }
        else:
            include_fields: StoryModelFieldsType = include

        include_fields["id"] = True

        if "user" in include_fields:
            if include_fields["user"] is True:
                include_fields["user"] = cast(
                    UserModelFieldsType, {key: True for key in get_fields(UserModel)}  # type: ignore
                )
            elif include_fields["user"] is False:
                include_fields["user"] = {"username": True}
            else:
                include_fields["user"]["username"] = True
        else:
            include_fields["user"] = {"username": True}

        data = cast(
            list[dict],
            await fetch_url(
                build_url(
                    f"stories/{self.data.id}/recommended",
                    fields=dict(include_fields),
                    limit=limit,
                    offset=offset,
                )
            ),
        )

        self.recommended = [Story(**story) for story in data]

        return data


# --- #


@singleton
class List:
    """A representation of a List on Wattpad.
    **Note**: Lists are singletons, unique as per their ID. Two List classes with the same ID are the _same_.

    Attributes:
        id (str): Lowercased ID of this List.
        name (str): The name of this List.
        user (User): The User who created this List.
        stories (list[Story]): Stories included within this List.
    """

    def __init__(self, id: str, user: User, name: str = "", stories: list[Story] = []):
        """Creates a List object.

        Args:
            id (str): The ID of this List.
            user (User): The User who created this List.
            name (str, optional): The name of this List. Defaults to "".
            stories (list[Story], optional): The Stories within this List. Defaults to [].
        """
        self.id = id.lower()
        self.name: str = name
        self.user: User = user
        self.stories: list[Story] = stories

    def __repr__(self) -> str:
        return f"<List id={self.id}>"



"""Copyright (C) 2024 TheOnlyWayUp

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program. If not, see https://www.gnu.org/licenses/.

---

Utility functions for the wattpad package."""

from typing import Any, Optional
import aiohttp
from aiohttp_client_cache.session import CachedSession
from pydantic import BaseModel
from os import environ

base_headers = {
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36 OPR/105.0.0.0"
}


def get_fields(model: BaseModel) -> list[str]:
    """Retrieve the fields of a Pydantic Model. Prefer field aliases if present.

    Args:
        model (BaseModel): The model to retrieve fields from.

    Returns:
        list[str]: A list of fields.
    """
    attribs = []
    for name, field in model.model_fields.items():
        if field.alias:
            attribs.append(field.alias)
        else:
            attribs.append(name)
    return attribs


def construct_fields(fields: dict) -> str:
    """Constructs a field query string from a dictionary representing the same.

    Example:
    ```py
    >>> d: StoryModelFieldsType = {
        "tags": True,
        "id": True,
        "parts": {"id": True},
        "tagRankings": True
    }
    >>> construct_fields(d)
    'tags,id,parts(id),tagRankings'
    ```

    Args:
        fields (dict): Field Data.

    Returns:
        str: Field Query String.
    """
    fields_str = ""

    for key, value in fields.items():
        if value is False:
            continue

        if value is True:
            fields_str += f"{key},"

        if type(value) is dict:
            fields_str += f"{key}({construct_fields(value)}),"

    fields_str = fields_str.removesuffix(",")

    return fields_str


def build_url(
    path: str,
    fields: Optional[dict] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> str:
    """Build an API Request URL.

    Args:
        path (str): The API Endpoint to request.
        fields (Optional[dict], optional): Fields Data, processed by `construct_fields`. Defaults to None.
        limit (Optional[int], optional): Number of records to limit the response to. Defaults to None.
        offset (Optional[int], optional): Number of records to skip before beginning the response. Defaults to None.

    Returns:
        str: The built URL.
    """
    base_url = f"https://www.wattpad.com/api/v3/{path}?"
    if fields:
        fields_str = construct_fields(fields)
        base_url += f"fields={fields_str}&"

    if limit:
        base_url += f"limit={limit}&"

    if offset:
        base_url += f"offset={offset}&"

    url = base_url.removesuffix("&")

    return url


async def fetch_url(url: str, headers: dict = {}) -> dict | list:
    """Perform a GET Request to the provided URL, merging the provided headers with `base_headers`.
    **Note**: API Responses are cached using the URL as a key. Set the `WPPY_SKIP_CACHE` Environment Variable to True to bypass the cache.

    Args:
        url (str): The URL to request.
        headers (dict, optional): Additional headers to merge atop of `base_headers`. Defaults to {}.

    Returns:
        dict | list: The JSON-Decoded Response.
    """
    use_headers = base_headers.copy()
    use_headers.update(headers)

    if environ.get("WPPY_SKIP_CACHE", False):
        session = aiohttp.ClientSession
    else:
        session = CachedSession

    async with session(headers=use_headers) as session:
        async with session.get(url) as response:
            response.raise_for_status()
            return await response.json()


def singleton(cls) -> Any:
    """Make a class a singleton using the first argument as the key.

    Thanks https://medium.com/@pavankumarmasters/exploring-the-singleton-design-pattern-in-python-a34efa5e8cfa.

    Returns:
        function: The nested `get_instance` function which handles singleton instance checking.
    """
    instances = {}

    def get_instance(*args, **kwargs):
        id = args[0].lower()

        if id in instances:
            return instances[id]
        else:
            instances[id] = cls(*args, **kwargs)

        return instances[id]

    return get_instance
